# wikirate4py
# Copyright 2021 Vasiliki Gkatziaki for WikiRate
# See LICENSE for details.

"""
wikirate4py WikiRate API library
"""
__version__ = '1.2.5'
__author__ = 'Vasiliki Gkatziaki'
__license__ = 'GPL-3.0'

from wikirate4py.api import API
from wikirate4py.cursor import Cursor
from wikirate4py.exceptions import (IllegalHttpMethod, WikiRate4PyException, HTTPException, BadRequestException,
                                    UnauthorizedException, ForbiddenException, NotFoundException,
                                    TooManyRequestsException,
                                    WikiRateServerErrorException)
from wikirate4py.mixins import WikiRateEntity
from wikirate4py.models import (Company, CompanyItem, Topic, TopicItem, Metric, MetricItem, ResearchGroup,
                                ResearchGroupItem, Project, ProjectItem, CompanyGroup, CompanyGroupItem, Source,
                                SourceItem, Answer, AnswerItem, RelationshipAnswer, RelationshipAnswerItem, Region,
                                Dataset, DatasetItem)
from wikirate4py.utils import to_dataframe



import json


class IllegalHttpMethod(Exception):
    """Exception raised when a request method is not supported by the api"""
    pass


class WikiRate4PyException(Exception):
    """Base exception for wikirate4py"""
    pass


class HTTPException(WikiRate4PyException):
    """Exception raised when an HTTP request fails"""

    def __init__(self, response):
        self.response = response

        self.api_errors = []
        self.api_codes = []
        self.api_messages = []

        try:
            response_json = response.json()
        except json.JSONDecodeError:
            super().__init__(f"{response.status_code} {response.reason}")
        else:
            errors = response_json.get("errors", {})
            error_text = "\t"
            for k, v in errors.items():
                error_text += k + ": " + str(v)

            super().__init__(
                f"{response.status_code} {response.reason}{error_text}"
            )


class BadRequestException(HTTPException):
    """Exception raised for a 400 HTTP status code"""
    pass


class UnauthorizedException(HTTPException):
    """Exception raised for a 401 HTTP status code"""
    pass


class ForbiddenException(HTTPException):
    """Exception raised for a 403 HTTP status code"""
    pass


class NotFoundException(HTTPException):
    """Exception raised for a 404 HTTP status code"""
    pass


class TooManyRequestsException(HTTPException):
    """Exception raised for a 429 HTTP status code"""
    pass


class WikiRateServerErrorException(HTTPException):
    """Exception raised for a 5xx HTTP status code"""
    pass



import time

import wikirate4py
from pymongo import MongoClient

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("fti_project")

api = wikirate4py.API('rMrVM3sSrtsiX23fKfsM5Qtt')

answers = db.pc_answers_2023.find({})

for answer in answers:
    # sources = set(answer['wikirate_sources'])
    # source = ''
    # for s in sources:
    #     source += s + "\n"
    #
    # value = ''
    # if len(answer['value']) > 1:
    #     for v in answer['value']:
    #         value += v + "\n"
    # else:
    #     value = answer['value'][0]

    metric_name = answer['metric'].split('+')[1]
    metric_designer = answer['metric'].split('+')[0]
    # comment = ''
    # if answer.get('comment') is not None:
    #     comment = answer['comment']

    for wid in answer['wikirate_id']:
        start_time = time.time()
        try:
            # if 'import_status_' + str(wid) in answer:
            #     continue
            print(answer['year'])
            print(metric_name)
            print(wid)
            a = api.update_research_metric_answer(metric_designer=metric_designer,
                                               metric_name=metric_name,
                                               company=wid,
                                               unpublished="0",
                                               year=answer['year'])
            end_time = time.time()
            exec_time = end_time - start_time
            print("Exec Time: " + exec_time.__str__())
            print(a)
            db.pc_answers_2023.update_one({'_id': answer['_id']}, {'$set': {'import_status_' + str(wid): 'imported'}})
        except Exception as e:
            print(e.__str__())

client.close()



import time
import csv
from datetime import datetime

import wikirate4py
from wikirate4py import Cursor

api = wikirate4py.API('ThessaloWikiRate')

company_groups = ['MSA_Renewable_Energy']
for company_group in company_groups:
    cursor = Cursor(api.get_answers_by_id, identifier=12641428, per_page=1000, company_group=company_group,
                    view="detailed")

    answers = []

    start_time = time.time()
    while cursor.has_next():
        answers.extend(cursor.next())
        end_time = time.time()
        start_time = time.time()

    end_time = time.time()
    exec_time = end_time - start_time
    print('Retrieving Dataset for {0} sector completed'.format(company_group))
    print(exec_time)
    print('------------------------------------------------------------------------------------')

    header = ["Answer Page", "Metric", "Company", "Year", "Value", "Source Page"]
    now = datetime.utcnow()
    date_time = now.strftime("%Y-%m-%d %H:%M:%S")

    rows = []

    for answer in answers:
        sources = ""
        for source in answer.sources:
            sources += 'https://wikirate.org/' + source + ", "
        if len(sources) > 0:
            sources = sources[:-2]
        row = [answer.url.replace('\.json', ''), answer.metric, answer.company, answer.year, answer.value, sources]
        rows.append(row)

    with open("/srv/msa-beyond-compliance/src/assets/downloads/Wikirate_MSA_Beyond_Compliance_Dashboard_Dataset_{0}.csv".format(
            'full' if company_group == '' else company_group), 'w',
            encoding='UTF8', newline='') as f:
        writer = csv.writer(f)
        writer.writerow([
            "# https://wikirate.org/MSA_Beyond_Compliance_Dashboard?filter[company_group]={0}".format(
                company_group)])
        writer.writerow(["# Creative Commons Attribution-ShareAlike 4.0 International License"])
        writer.writerow(["# Last updated at: {0} UTC".format(date_time)])
        writer.writerow(["# "])
        writer.writerow(header)
        # write the data
        writer.writerows(rows)

    f.close()



import pandas as pd
from pymongo.errors import DuplicateKeyError

import wikirate4py
from pymongo import MongoClient

from wikirate4py import Cursor

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("msa_assessed")

api = wikirate4py.API('ThessaloWikiRate')

assessed_metrics = ["Business & Human Rights Resource Centre+Modern Slavery Statement",
                    "Business & Human Rights Resource Centre+MSA statement signed",
                    "Business & Human Rights Resource Centre+MSA Statement Approval",
                    "Walk Free+MSA Organizational structure and operations",
                    "Walk Free+MSA supply chain disclosure",
                    "Walk Free+MSA policy (revised)",
                    "Walk Free+MSA risk assessment",
                    "Walk Free+MSA Identification of risks",
                    "Walk Free+MSA risk management (revised)",
                    "Walk Free+MSA whistleblowing mechanism (revised)",
                    "Walk Free+MSA incidents identified",
                    "Walk Free+MSA incidents remediation (revised)",
                    "Walk Free+MSA training (revised)",
                    "Walk Free+MSA Performance Indicators",
                    "Walk Free+MSA Business Performance Indicators"]


def get_uk_assessed_statements():
    cursor = Cursor(api.get_answers_by_metric_id, metric_id=12602618, value='Yes', per_page=100)

    while cursor.has_next():
        answers = cursor.next()

        for answer in answers:
            print(answer)
            db.uk_assessed_0523.insert_one(answer.json())


def get_answers_on_assessed_metrics():
    assessed_statements = db.uk_assessed_0523.find()

    for statement in assessed_statements:
        # print(statement)
        for metric in assessed_metrics:
            metric_designer = metric.split('+')[0]
            metric_name = metric.split('+')[1]
            answer = api.get_answer_by(metric_designer,
                                       metric_name,
                                       statement['company'],
                                       statement['year'])

            json = answer.raw_json()
            json['_id'] = json['id']
            del json['id']
            try:
                db.uk_assessed_statements.insert_one(json)
            except DuplicateKeyError as e:
                print('duplicate')


def verify_statements():
    assessed_statements = db.uk_assessed_0523.find()

    for statement in assessed_statements:
        verified = True
        number_of_verified_answers = 0

        verified_answers = []
        verified_answers_ids = []
        unverified_answers = []
        unverified_answers_ids = []
        for metric in assessed_metrics:
            answer = db.uk_assessed_statements.find_one(
                {'metric': metric, 'company': statement['company'], 'year': statement['year']})
            print(answer)
            if len(answer['checked_by']) == 0:
                verified = False
                unverified_answers.append(answer['metric'])
                unverified_answers_ids.append(answer['_id'])
            else:
                number_of_verified_answers += 1
                verified_answers.append(answer['metric'])
                verified_answers_ids.append(answer['_id'])
        if verified:
            db.uk_assessed_0523.update_one({'_id': statement['_id']},
                                           {'$set': {'verified': True,
                                                     'verified_answers': verified_answers,
                                                     'unverified_answers': unverified_answers,
                                                     'verified_answers_ids': verified_answers_ids,
                                                     'unverified_answers_ids': unverified_answers_ids
                                                     }})
        else:
            db.uk_assessed_0523.update_one({'_id': statement['_id']},
                                           {'$set': {'number_of_verified_answers': number_of_verified_answers,
                                                     'verified_answers': verified_answers,
                                                     'unverified_answers': unverified_answers,
                                                     'verified_answers_ids': verified_answers_ids,
                                                     'unverified_answers_ids': unverified_answers_ids
                                                     }})


assessed_statements = db.uk_assessed_0523.find({'verified': {'$exists': False}})

f = open("D:\\Assessed Stats.csv", "a")

for statement in assessed_statements:
    f.write("{0};{1};{2};{3};{4};{5};{6}\n".format(
        statement['company'],
        statement['year'],
        statement['number_of_verified_answers'],
        statement['unverified_answers'],
        statement['verified_answers'],
        statement['unverified_answers_ids'],
        statement['verified_answers_ids']
    ))

f.close()

# get_uk_assessed_statements()
# get_answers_on_assessed_metrics()



import pandas as pd
import wikirate4py
from pymongo import MongoClient

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("sec_filings")

df = pd.read_excel('D:/WikiRate/data/sec/Selected SEC metrics.xlsx',
                   sheet_name='metrics', engine='openpyxl')

api = wikirate4py.API('ThessaloWikiRate')

for index, row in df.iterrows():
    designer = row['metric_designer']
    name = row['metric_name']
    tag = row['tag']
    about = row['about']
    value_type = row['value_type']
    unit = row['unit']
    question = row['question']

    print(index)
    print(name)
    try:

        metric = api.add_metric(designer=designer,
                                name=name,
                                question=question,
                                topics=['Annual Reporting', 'Corporate Accountability', 'Financial'],
                                metric_type='Researched',
                                value_type=value_type,
                                unit='USD',
                                research_policy='Community Assessed',
                                report_type='Form 10-K',
                                about=about,
                                methodology='<p>Companies should report their financial performance in their annual financial reports, and these are the best places to find this information.</p><p>For companies that file with the SEC, this information can be found within their \"Form 10-K\" filings. You can often find these documents by searching for \"Company 10-k\", or they can be found through the <a href=\"http://www.sec.gov/edgar/searchedgar/companysearch.html\">SEC\'s EDGAR search</a>.</p>')

        json = metric.json()
        json['_id'] = json['id']
        del json['about']
        del json['methodology']

        del json['id']
        json['tag'] = row['tag']
        db.metrics.insert_one(json)

        print(json)
    except Exception as e:
        print(e.__str__())



import wikirate4py

topics = ["UNGP 16", "UNGPRF A2", "CHRB A.2.1", "KTC 2.1", "ISO 30414 4.7.5", "OECD Principle 2", "PRI Principle 1",
          "UNGP 12", "UNGP 15", "UNGPRF A1", "UNGPRF C1", "GRI 106", "GRI 408: Child Labour", "DJSI 3.3.1",
          "CHRB A.1.1", "CHRB D.1.5/D.2.5", "CWC 1.2", "CWC 1.2.1", "CWC 4.2", "KTC 2.2", "SDG 8.7", "SDG 16.2",
          "ILO 105", "ILO 29", "SFDR Social metrics", "OECD Principle 4", "OECD Principle 5", "PRI Principle 3",
          "UNGC 1", "UNGC 4", "WBA CSI 1", "WBA CSI 2", "UNGP 17", "UNGP 18", "UNGPRF C6", "DJSI 3.3.2", "CHRB B.2.1",
          "CHRB B.2.2", "CHRB B.1.3/B.2.3", "KTC 5.1", "ETI RF B4.2", "ISO26000 Clause 6.3.3", "UNDHR Article 4",
          "GRI 409: Forced or Compulsory Labor", "ISO 30414 4.7.12", "CWC 1.1", "CWC 1.1.1", "CWC 1.1.3", "CWC 1.5",
          "WBA CSI 13", "DJSI 3.5.6", "WGEA Question 1", "UNDHR Article 25", "ILO 111", "UNGC 6",
          "ISO26000 Clause 6.3.7", "UNDHR Article 2", "UNDHR Article 7", "UNDHR Article 23", "UNGP 21",
          "GRI 406: Non-discrimination", "CHRB D.1.1a/D.2.1a/D.3.1", "CHRB D.1.1b/D.2.1b", "CWC 6.1", "SDG 8.5",
          "SDG 10.4", "GRI 404: Training and Education", "SDG 5.1", "SDG 5.5", "SFDR Social Metrics",
          "GRI 404: Training and Education", "GRI 405: Diversity and Equal Opportunity", "CWC 4.7.1", "SDG 16.7",
          "CHRB D.1.7.a/D.2.7a", "CWC 5.1", "CWC 5.4", "CWC 5.4.1", "ISO 30414 4.7.7", "CHRB D.1.7.b/D.2.7.b",
          "CWC 5.4.2", "UNGP 13", "UNGP 19", "DJSI 3.2.6", "CWC 2.1", "WBA CSI 12", "SDG 8.8", "CHRB C.1", "CWC 7.3",
          "WBA CSI 8", "ISO26000 Clause 6.3.1", "UNGP 31", "ISO 30414 4.7.2", "CHRB D.1.3", "CWC 4.1", "CWC 4.1.2",
          "KTC 4.2", "CHRB B.1.4B", "CHRB B.1.7", "CWC 4.5", "UNGPRF A1.3", "DJSI 3.6.1", "CWC 4.3", "KTC 1.2",
          "SDG 10.6", "ETI RF B1.1", "ETI RF B3.2", "KTC 9.1"]

api = wikirate4py.API('ThessaloWikiRate')
for topic in topics:
    try:
        topic = api.add_topic(topic)
        print(topic)
    except Exception as ex:
        print(ex.__str__())



import wikirate4py
from wikirate4py import Cursor

api = wikirate4py.API('ThessaloWikiRate')

# cursor = Cursor(api.get_answers_by_id, identifier=3233883, source="Source-000162318", per_page=100)
# answers = []
# while cursor.has_next():
#     answers.extend(cursor.next())
#
# print(len(answers))
# for answer in answers:
#     print(answer)
#     api.delete_wikirate_entity(id=answer.id)

metrics = [6908731, 8408492, 14019318, 14019304, 13006160, 8396541, 14019252]
# metrics = []
# metrics = []
for metric in metrics:
    print(metric)
    api.delete_wikirate_entity(id=metric)


# metrics = [13006134, 8412610, 14019233]

# for metric in metrics:
#     cursor = Cursor(api.get_answers_by_id, identifier=metric, per_page=100)
#     while cursor.has_next():
#         answers = cursor.next()
#         for answer in answers:
#             print(answer.id)
#             try:
#                 a = api.update_research_metric_answer_by_id(
#                     id=answer.id,
#                     metric_name='MSA_State_imposed_forced_labour',
#                     metric_designer='Walk Free',
#                     year=answer.year,
#                     company=answer.company)
#             except Exception as e:
#                 print(e)



import math

import pandas as pd
from pymongo.errors import DuplicateKeyError

import wikirate4py
from pymongo import MongoClient

from wikirate4py import Cursor

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("msa_assessed")

api = wikirate4py.API('ThessaloWikiRate')

df = pd.read_excel("C:\\Users\\vasgat\\Downloads\\TEST MSA research copy to subsidiaries(1).xlsx",
                   sheet_name='CompanyYears', engine='openpyxl')

api = wikirate4py.API('ThessaloWikiRate')

for index, row in df.iterrows():
    company = row['Company name']
    year = row['Year']
    id = row['Company ID']
    c = None
    if math.isnan(id):
        try:
            c = api.get_company(company)
            id = c.id
        except Exception as e:
            print("Nan: " + company)
    print(id)

    if not math.isnan(id):
        if c is None:
            c = api.get_company(int(id))
        assessment = db.uk_assessed_0523.find_one({'company': c.name, 'year': year})
        if assessment is not None:
            db.uk_assessed_0523.update_one({'_id': assessment['_id']}, {'$set': {'wikirate_id': int(id)}})



from pymongo import MongoClient

import wikirate4py
from wikirate4py import Cursor

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("msa_project")

api = wikirate4py.API('ThessaloWikiRate')

# cursor = Cursor(api.get_relationship_answers, metric_name='Subsidiary Of', metric_designer='Commons', view='detailed',
#                 per_page=200)
#
# while cursor.has_next():
#     subsidiaries = cursor.next()
#     for s in subsidiaries:
#         json = s.json()
#         json['_id'] = json['id']
#         del json['id']
#         db.platform_subsidiaries.insert_one(json)

subsidiaries = db.platform_subsidiaries.find({'sources': {'$in': ['Source-000139335', 'Source-000127271']}})

for subsidiary in subsidiaries:
    print(subsidiary['_id'])
    api.delete_wikirate_entity(subsidiary['_id'])

client.close()



import time
import csv
from datetime import datetime

import wikirate4py
from wikirate4py import Cursor

api = wikirate4py.API('ThessaloWikiRate')


companies = []
# companies = api.get_content('~12641428+company')
#
# print(companies)
# print(len(companies))
#
metrics = []

cursor = Cursor(api.get_metrics_of, identifier=12641428, per_page=200)
while cursor.has_next():
    m_set = cursor.next()
    for metric in m_set:
        metrics.append(metric.id)
    break

print(metrics)
print(len(metrics))

company_groups = ['MSA_Renewable_Energy']
for company_group in company_groups:
    rows = []
    answers = []
    for metric in metrics:
        print(metric)
        cursor = Cursor(api.get_answers_by_id, identifier=metric, per_page=100, company_group=company_group,
                        view="detailed")

        while cursor.has_next():
            answers.extend(cursor.next())

    header = ["Answer Page", "Metric", "Company", "Year", "Value", "Source Page"]
    now = datetime.utcnow()
    date_time = now.strftime("%Y-%m-%d %H:%M:%S")

    for answer in answers:
        if company_group == '' and answer.company in companies:
            sources = ""
            for source in answer.sources:
                sources += 'https://wikirate.org/' + source + ", "
            if len(sources) > 0:
                sources = sources[:-2]
            row = [answer.url.replace('\.json', ''), answer.metric, answer.company, answer.year, answer.value,
                   sources]
            rows.append(row)
        elif company_group != '':
            sources = ""
            for source in answer.sources:
                sources += 'https://wikirate.org/' + source + ", "
            if len(sources) > 0:
                sources = sources[:-2]
            row = [answer.url.replace('\.json', ''), answer.metric, answer.company, answer.year, answer.value,
                   sources]
            rows.append(row)
    with open(
            "D:\\Wikirate_MSA_Beyond_Compliance_Dashboard_Dataset_{0}.csv".format(
                'full' if company_group == '' else company_group), 'w',
            encoding='UTF8', newline='') as f:
        writer = csv.writer(f)
        writer.writerow([
            "# https://wikirate.org/MSA_Beyond_Compliance_Dashboard?filter[company_group]={0}".format(
                company_group)])
        writer.writerow(["# Creative Commons Attribution-ShareAlike 4.0 International License"])
        writer.writerow(["# Last updated at: {0} UTC".format(date_time)])
        writer.writerow(["# "])
        writer.writerow(header)
        # write the data
        writer.writerows(rows)

    f.close()



from typing import List

from bson import ObjectId
from pymongo import MongoClient
from wikirate4py.models import SourceItem

from wikirate4py.exceptions import NotFoundException

import wikirate4py

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("msa_assessed")

api = wikirate4py.API('NzSK5muu3o7oSwDa2PpUVQtt')


def sources_to_list(sources):
    list_of_sources = []
    for s in sources:
        if isinstance(s, SourceItem):
            list_of_sources.append(s.name)
        else:
            list_of_sources.append(s['name'])
    return list_of_sources


def conflict_free(verified_answers_ids, entity):
    conflict_free = True
    for id in verified_answers_ids:
        answer = db.aus_assessed_statements.find_one({'_id': id})
        metric = answer['name'].split('+')
        try:
            a = api.get_answer_by(metric[0], metric[1], entity, metric[3])
            original_sources = sources_to_list(answer['sources'])
            child_sources = sources_to_list(a.sources)

            if answer['value'] == a.value and any(item in child_sources for item in original_sources):
                continue
            else:
                print(answer)
                print(a)
                print()
                conflict_free = False
        except NotFoundException as e:
            continue
    return conflict_free


statements = db.assessed_0523.find({"verified": True, "related_entities": {'$exists': True}})

counter = 0
for statement in statements:
    print('Statement ID: {0}'.format(statement['_id']))
    related_entities = statement['related_entities']
    verified_answers_ids = statement['verified_answers_ids']

    for entity in related_entities:
        print(entity)
        is_conflict_free = conflict_free(verified_answers_ids, entity)
        if is_conflict_free:
            for id in verified_answers_ids:
                answer = db.aus_assessed_statements.find_one({'_id': id})
                metric = answer['name'].split('+')
                try:
                    a = api.get_answer_by(metric[0], metric[1], entity, metric[3])
                    # api.verify_answer(a.id)
                    # if a.metric == 'Business & Human Rights Resource Centre+Modern Slavery Statement':
                    #     api.verify_answer(a.id)
                    #     print(a)
                    #     continue
                    #
                    # source = ''
                    # for s in sources_to_list(answer['sources']):
                    #     source += s + '\n'
                    #
                    # value = ''
                    # if answer['value'].__contains__(', '):
                    #     for v in answer['value'].split(', '):
                    #         value += v + '\n'
                    # else:
                    #     value = answer['value']
                    # try:
                    #     comments = api.get_comments(answer['_id'])
                    # except NotFoundException as ex:
                    #     comments = ''
                    # copied_answer = api.update_research_metric_answer(
                    #     metric_designer=metric[0],
                    #     metric_name=metric[1],
                    #     company=entity,
                    #     value=value,
                    #     year=answer['year'],
                    #     source=source,
                    #     comment=comments + '<hr>' + ' Data copied from <b>' + answer[
                    #         'company'] + ' for ' +
                    #             answer['year'].__str__() + '</b>'
                    # )
                    # print(copied_answer)
                    # api.verify_answer(copied_answer.id)
                except NotFoundException as e:
                    source = ''
                    for s in sources_to_list(answer['sources']):
                        source += s + '\n'

                    value = ''
                    if answer['value'].__contains__(', '):
                        for v in answer['value'].split(', '):
                            value += v + '\n'
                    else:
                        value = answer['value']
                    try:
                        comments = api.get_comments(answer['_id'])
                    except NotFoundException as ex:
                        comments = ''
                    copied_answer = api.update_research_metric_answer(
                        metric_designer=metric[0],
                        metric_name=metric[1],
                        company=entity,
                        value=value,
                        year=answer['year'],
                        source=source,
                        comment=comments + '<hr>' + ' Data copied from <b>' + answer[
                            'company'] + ' for ' +
                                answer['year'].__str__() + '</b>'
                    )
                    print(copied_answer)
                    api.verify_answer(copied_answer.id)

            db.assessed_0523.update_one({'_id': statement['_id']},
                                        {'$pull': {'related_entities': entity},
                                         '$push': {'answers_copied_to_rc': entity}})

print(counter)



from urllib.parse import quote

import wikirate4py
import urllib.request

from wikirate4py import Cursor

api = wikirate4py.API('NzSK5muu3o7oSwDa2PpUVQtt')

cursor = Cursor(api.get_sources,
                wikirate_link='downloads.modern-slavery-statement-registry.service.gov.uk/pdf-published/', per_page=100)
while cursor.has_next():
    sources = cursor.next()
    for source in sources:
        if source.file_url == '':
            try:
                print(source)
                result = urllib.parse.urlparse(source.original_source)
                url = "https://{0}{1}".format(result.netloc, quote(result.path))
                urllib.request.urlretrieve(url, 'D://modern-slavery-statement.pdf')
                api.upload_file(source=source.name, file='D://modern-slavery-statement.pdf')
            except Exception as e:
                print(e.__str__())



class WikiRateEntity(object):
    __slots__ = ()

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id
        return NotImplemented

    def json(self):
        return {key: getattr(self, key, None) for key in self.__slots__ if key != "raw"}

    def raw_json(self):
        return self.raw

    def get_parameters(self):
        return self.__slots__

    def __repr__(self):
        return str(self.json())



import wikirate4py

api = wikirate4py.API('ThessaloWikiRate')

answers = api.get_relationship_answers(metric_name='Supplier of',
                                            metric_designer='Commons',
                                            country='United Kingdom')


import wikirate4py
from pymongo import MongoClient

from wikirate4py import Cursor

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("msa_assessed")

api = wikirate4py.API('ThessaloWikiRate')

statements = db.assessed_0523.find({'verified': True, 'related_entities': {'$exists': False}})

for statement in statements:
    s = db.aus_assessed_statements.find_one({'wikirate_id': statement['wikirate_id'],
                                            'year': statement['year'],
                                            'metric': 'Business & Human Rights Resource Centre+Modern Slavery Statement'})

    if s is not None and len(s['sources']) == 1:
        answers = api.get_answers_by_id(s['sources'][0]['id'], metric_name='Modern Slavery Statement',
                                        year=statement['year'])
        if len(answers) > 1:
            print(s['sources'][0]['name'])
            print(statement['company'])
            for answer in answers:
                print(answer.company)
            print()
    elif s is not None and len(s['sources']) > 1:
        for source in s['sources']:
            answers = api.get_answers_by_id(source['id'], metric_name='Modern Slavery Statement')
            if len(answers) > 1:
                print(source['name'])
                print(statement['company'])
                for answer in answers:
                    print(answer.company)
                print()

client.close()



from pandas import DataFrame

from wikirate4py.exceptions import WikiRate4PyException
from wikirate4py.mixins import WikiRateEntity
import html2text


class Company(WikiRateEntity):
    __slots__ = ("id", "name", "headquarters", "wikipedia_url",
                 "aliases", "url", "oar_id", "sec_cik", "open_corporates", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 651:
            raise WikiRate4PyException('Invalid type of entity')
        self.id = data.get("id")
        self.name = data["name"]
        self.headquarters = None
        if data.get("headquarters").get("content") is not None:
            self.headquarters = data.get("headquarters", None).get("content", None)[0]
        self.aliases = data.get("alias").get("content")
        self.url = data.get("html_url")
        self.wikipedia_url = data.get("wikipedia").get("content")
        self.oar_id = data.get("oar_id").get("content")
        self.sec_cik = data.get("sec_cik").get("content")
        self.open_corporates = data.get("open_corporates").get("content")


class CompanyItem(WikiRateEntity):
    __slots__ = ("id", "name", "headquarters", "wikipedia_url",
                 "aliases", "url", "oar_id", "sec_cik", "open_corporates", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != "Company":
            raise WikiRate4PyException('Invalid type of entity')
        self.id = data.get("id")
        self.name = data["name"]
        self.headquarters = data.get("headquarters")
        self.aliases = data.get("alias")
        self.url = data.get("url").replace(".json", "")
        self.wikipedia_url = data.get("wikipedia")
        self.oar_id = data.get("oar_id")
        self.sec_cik = data.get("sec_cik")
        self.open_corporates = data.get("open_corporates")


class Topic(WikiRateEntity):
    __slots__ = ("id", "name", "metrics", "projects", "bookmarkers", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 1010:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.metrics = data.get("metrics", 0)
        self.projects = data.get("projects", 0)
        self.bookmarkers = data.get("bookmarkers", 0)
        self.url = data.get("html_url")


class Project(WikiRateEntity):
    __slots__ = ("id", "name", "metrics", "companies", "answers", "created_at", "updated_at", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 39830:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.metrics = data.get("metrics", {}).get("content", [])
        self.companies = data.get("companies", {}).get("content", [])
        self.answers = [AnswerItem(item) for item in data.get("items", {})]
        self.created_at = data.get("created_at")
        self.updated_at = data.get("updated_at")
        self.url = data.get("html_url")

    def to_dataframe(self):
        answers = []
        for answer in self.answers:
            answers.append(answer.json())
        return DataFrame.from_dict(answers)


class ProjectItem(WikiRateEntity):
    __slots__ = ("id", "name", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Project':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.url = data.get("url").replace(".json", "")


class Dataset(WikiRateEntity):
    __slots__ = ("id", "name", "metrics", "companies", "answers", "license", "created_at", "updated_at", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 7926098:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.metrics = data.get("metrics", {}).get("content", [])
        self.companies = data.get("companies", {}).get("content", [])
        self.answers = [AnswerItem(item) for item in data.get("items", {})]
        self.license = data.get("license")
        self.created_at = data.get("created_at")
        self.updated_at = data.get("updated_at")
        self.url = data.get("html_url")

    def to_dataframe(self):
        answers = []
        for answer in self.answers:
            answers.append(answer.json())
        return DataFrame.from_dict(answers)


class DatasetItem(WikiRateEntity):
    __slots__ = ("id", "name", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Data Set':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.url = data.get("url").replace(".json", "")


class TopicItem(WikiRateEntity):
    __slots__ = ("id", "name", "metrics", "projects", "bookmarkers", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Topic':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.metrics = data.get("metrics", 0)
        self.projects = data.get("projects", 0)
        self.bookmarkers = data.get("bookmarkers", 0)
        self.url = data.get("url").replace(".json", "")


class Metric(WikiRateEntity):
    __slots__ = (
        "id", "name", "designer", "question", "metric_type", "about", "methodology", "value_type",
        "value_options", "report_type", "research_policy", "unit", "range", "hybrid", "topics", "scores",
        "formula", "answers", "bookmarkers", "projects", "calculations", "answers_url", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 43576:
            raise WikiRate4PyException('Invalid type of entity')
        h = html2text.HTML2Text()

        self.id = data.get("id")
        self.designer = data["designer"]
        self.name = data["title"]
        self.question = data.get("question", {}).get("content")
        self.about = h.handle(data.get("about").get("content", ""))
        self.methodology = h.handle(data.get("methodology").get("content", ""))
        self.value_type = data.get("value_type").get("content")
        self.value_options = data.get("value_options", {}).get("content", [])
        if len(self.value_options) == 1 and self.value_options[0] == "Unknown":
            self.value_options = []
        self.report_type = data.get("report_type").get("content")
        self.metric_type = data.get("metric_type").get("content")
        self.research_policy = data.get("research_policy").get("content")
        self.unit = data.get("unit").get("content")
        self.range = data.get("range").get("content")
        self.hybrid = data.get("hybrid").get("content")
        self.topics = data.get("topics").get("content")
        self.scores = data.get("scores").get("content")
        self.formula = data.get("formula").get("formula")
        self.answers = data.get("answers")
        self.bookmarkers = data.get("bookmarkers")
        self.projects = data.get("projects")
        self.calculations = data.get("calculations")
        self.answers_url = data.get("answers_url")

        self.url = data.get("html_url")


class MetricItem(WikiRateEntity):
    __slots__ = (
        "id", "name", "designer", "question", "metric_type", "about", "methodology", "value_type",
        "value_options", "report_type", "research_policy", "unit", "range", "hybrid", "topics", "scores",
        "formula", "answers", "bookmarkers", "projects", "calculations", "answers_url", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Metric':
            raise WikiRate4PyException('Invalid type of entity')
        h = html2text.HTML2Text()

        self.id = data.get("id")
        self.designer = data["designer"]
        self.name = data["title"]
        self.question = data.get("question")
        if data.get("about") is not None:
            self.about = h.handle(data.get("about", " "))
        if data.get("methodology") is not None:
            self.methodology = h.handle(data.get("methodology", ""))
        self.value_type = data.get("value_type")
        self.value_options = data.get("value_options")
        if len(self.value_options) == 1 and self.value_options[0] == "Unknown":
            self.value_options = None
        self.report_type = data.get("report_type")
        self.metric_type = data.get("metric_type")
        self.research_policy = data.get("research_policy")
        self.unit = data.get("unit")
        self.range = data.get("range")
        self.hybrid = data.get("hybrid")
        self.topics = data.get("topics")
        self.scores = data.get("scores")
        self.formula = data.get("formula")
        self.answers = data.get("answers")
        self.bookmarkers = data.get("bookmarkers")
        self.projects = data.get("projects")
        self.calculations = data.get("calculations")
        self.answers_url = data.get("answers_url")

        self.url = data.get("url").replace(".json", "")


class ResearchGroup(WikiRateEntity):
    __slots__ = (
        "id", "name", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 2301582:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.url = data.get("html_url")


class ResearchGroupItem(WikiRateEntity):
    __slots__ = (
        "id", "name", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Research Group':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.url = data.get("url").replace(".json", "")


class CompanyGroup(WikiRateEntity):
    __slots__ = (
        "id", "name", "url", "members", "members_links", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 5458825:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.members = data.get("companies", {}).get("content", [])
        self.members_links = data["links"]
        self.url = data.get("html_url")


class CompanyGroupItem(WikiRateEntity):
    __slots__ = (
        "id", "name", "url", "members", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Company Group':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data["name"]
        self.members = data.get("companies", [])
        self.url = data.get("url").replace('.json', '')


class Source(WikiRateEntity):
    __slots__ = (
        "id", "name", "title", "description", "url", "original_source", "file_url", "year", "metrics", "companies",
        "report_type", "created_at", "updated_at", "raw")

    def __init__(self, data):
        self.raw = data
        if data.get("type").get("id") != 629:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data.get("name")
        self.title = data.get("title", {}).get("content")
        self.description = data.get("description", {}).get("content")
        self.file_url = data.get("file", {}).get("content")
        self.original_source = data.get("link", {}).get("content")
        self.year = data.get("year", {}).get("content")
        self.report_type = data.get("report_type", {}).get("content")
        self.metrics = data.get("metric", {}).get("content")
        self.companies = data.get("company", {}).get("content")
        self.created_at = data.get("created_at")
        self.updated_at = data.get("updated_at")
        self.url = data.get("html_url")


class SourceItem(WikiRateEntity):
    __slots__ = (
        "id", "name", "title", "url", "original_source", "file_url", "year", "report_type", "raw")

    def __init__(self, data):
        self.raw = data
        if data.get("type") != 'Source':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data.get("name")
        self.title = data.get("title")
        self.file_url = data.get("file")
        self.original_source = data.get("link")
        self.year = data.get("year")
        self.report_type = data.get("report_type")
        self.url = data.get("url").replace(".json", "")


class Answer(WikiRateEntity):
    __slots__ = (
        "id", "metric", "company", "value", "year", "comments", "sources", "checked_by",
        "check_requested", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data.get("type") != "Answer" and data.get("type", {}).get("id") != 43678:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.metric = data["metric"]
        self.company = data["company"]
        self.value = data.get("value")
        self.year = data.get("year")
        self.comments = data.get("comments")
        self.sources = []
        if not data.get("sources").__str__().__contains__("Error rendering:"):
            for s in data.get("sources", []):
                self.sources.append(SourceItem(s))
        self.checked_by = data.get("checked_by", [])
        # self.check_requested = data.get("checked_by", {}).get("check_requested")
        self.url = data.get("html_url")


class AnswerItem(WikiRateEntity):
    __slots__ = (
        "id", "metric", "company", "value", "year", "comments", "sources", "checked_by",
        "check_requested", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Answer':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.metric = data["metric"]
        self.company = data["company"]
        self.value = data.get("value")
        self.year = data.get("year")
        self.comments = data.get("comments")
        self.sources = [SourceItem(item) for item in data.get("sources", {})]
        self.url = data.get("url").replace(".json", "")


class RelationshipAnswer(WikiRateEntity):
    __slots__ = (
        "id", "metric", "value", "year", "comments", "sources", "checked_by",
        "subject_company_name", "subject_company_id", "object_company_name", "object_company_id", "check_requested",
        "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 2534606:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        fields = data.get("name").split("+")
        self.metric = fields[0] + "+" + fields[1]
        self.value = data.get("value")
        self.year = data.get("year")
        self.comments = data.get("comments")
        self.sources = []
        for s in data.get("sources", []):
            self.sources.append(s.get("name"))
        if data.get("checked_by") is not None:
            self.checked_by = data.get("checked_by").get("content")
            self.check_requested = data.get("checked_by").get("check_requested")
        self.subject_company_name = data.get("subject_company").get("name")
        self.object_company_name = data.get("object_company").get("name")
        self.subject_company_id = data.get("subject_company").get("id")
        self.object_company_id = data.get("object_company").get("id")
        self.url = data.get("html_url")


class RelationshipAnswerItem(WikiRateEntity):
    __slots__ = (
        "id", "metric", "metric_id", "value", "year", "comments", "sources",
        "subject_company_name", "object_company_name", "subject_company_id", "object_company_id",
        "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Relationship Answer' and data.get("type", {}).get("name") != 'Relationship Answer':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        fields = data.get("name").split("+")
        self.metric = fields[0] + "+" + fields[1]
        self.metric_id = data.get("metric_id")
        self.value = data.get("value")
        self.year = data.get("year")
        self.subject_company_id = data.get("subject_company_id")
        self.object_company_id = data.get("object_company_id")
        self.comments = data.get("comments")
        self.sources = []
        for s in data.get("sources", []):
            self.sources.append(s.get("name"))
        self.subject_company_name = data.get("subject_company") \
            if isinstance(data.get("subject_company"), str) else data.get("subject_company", {}).get("name")
        self.object_company_name = data.get("object_company") \
            if isinstance(data.get("object_company"), str) else data.get("object_company", {}).get("name")
        self.url = data.get("url").replace(".json", "")


class RegionItem(WikiRateEntity):
    __slots__ = (
        "id", "name", "url", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"] != 'Region':
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data.get("name")
        self.url = data.get("url").replace(".json", "")


class Region(WikiRateEntity):
    __slots__ = (
        "id", "name", "url", "oc_jurisdiction_key", "region", "country", "raw")

    def __init__(self, data):
        self.raw = data
        if data["type"]["id"] != 7044738:
            raise WikiRate4PyException('Invalid type of entity')

        self.id = data.get("id")
        self.name = data.get("name")
        self.url = data.get("url").replace(".json", "")
        self.oc_jurisdiction_key = data.get("items", [])[1].get("content")
        self.country = data.get("items", [])[3].get("content")
        self.region = data.get("items", [])[2].get("content")



import time

import wikirate4py
from wikirate4py import Cursor

api = wikirate4py.API('ThessaloWikiRate')

cursor = Cursor(api.get_answers_by_id,
                identifier=826615,
                year=2021,
                view='detailed',
                verification='community_verified'
                )

answers = []
while cursor.has_next():
    answers.extend(cursor.next())




import pandas as pd
from pymongo.errors import DuplicateKeyError

import wikirate4py
from pymongo import MongoClient

from wikirate4py import Cursor

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("msa_assessed")

api = wikirate4py.API('ThessaloWikiRate')

cursor = wikirate4py.Cursor(api.get_relationship_metric_answers,
                            metric_designer="Commons",
                            metric_name="Supplied_by",
                            object_company_id=3438606)

relationships = []

while cursor.has_next():
    relationships += cursor.next()
    print(len(relationships))
print(relationships)
#
#
# def get_aus_assessed_statements():
#     cursor = Cursor(api.get_answers_by_metric_id, metric_id=12602630, value='Yes', per_page=100)
#
#     while cursor.has_next():
#         answers = cursor.next()
#
#         for answer in answers:
#             print(answer)
#             db.assessed_0523.insert_one(answer.json())
#
#
# def get_answers_on_assessed_metrics():
#     assessed_statements = db.assessed_0523.find()
#
#     assessed_metrics = [
#         "Business & Human Rights Resource Centre+MSA statement signed",
#         "Business & Human Rights Resource Centre+MSA Statement Approval",
#         "Walk Free+MSA Organizational structure and operations",
#         "Walk Free+MSA supply chain disclosure",
#         "Walk Free+MSA policy (revised)",
#         "Walk Free+MSA risk assessment",
#         "Walk Free+MSA Identification of risks",
#         "Walk Free+MSA risk management (revised)",
#         "Walk Free+MSA whistleblowing mechanism (revised)",
#         "Walk Free+MSA incidents identified",
#         "Walk Free+MSA incidents remediation (revised)",
#         "Walk Free+MSA training (revised)",
#         "Walk Free+MSA Performance Indicators",
#         "Walk Free+MSA Business Performance Indicators",
#         "Walk Free+MSA Consultation Process"]
#
#     for statement in assessed_statements:
#         # print(statement)
#         for metric in assessed_metrics:
#             metric_designer = metric.split('+')[0]
#             metric_name = metric.split('+')[1]
#             answer = api.get_answer_by(metric_designer,
#                                        metric_name,
#                                        statement['company'],
#                                        statement['year'])
#
#             json = answer.raw_json()
#             json['_id'] = json['id']
#             del json['id']
#             try:
#                 db.aus_assessed_statements.insert_one(json)
#             except DuplicateKeyError as e:
#                 print('duplicate')
#
#
# assessed_metrics = [
#     "Business & Human Rights Resource Centre+Modern Slavery Statement",
#     "Business & Human Rights Resource Centre+MSA statement signed",
#     "Business & Human Rights Resource Centre+MSA Statement Approval",
#     "Walk Free+MSA Organizational structure and operations",
#     "Walk Free+MSA supply chain disclosure",
#     "Walk Free+MSA policy (revised)",
#     "Walk Free+MSA risk assessment",
#     "Walk Free+MSA Identification of risks",
#     "Walk Free+MSA risk management (revised)",
#     "Walk Free+MSA whistleblowing mechanism (revised)",
#     "Walk Free+MSA incidents identified",
#     "Walk Free+MSA incidents remediation (revised)",
#     "Walk Free+MSA training (revised)",
#     "Walk Free+MSA Performance Indicators",
#     "Walk Free+MSA Business Performance Indicators",
#     "Walk Free+MSA Consultation Process"]
#
#
# def verify_statements():
#     assessed_statements = db.assessed_0523.find()
#
#     for statement in assessed_statements:
#         verified = True
#         number_of_verified_answers = 0
#
#         verified_answers = []
#         verified_answers_ids = []
#         unverified_answers = []
#         unverified_answers_ids = []
#         for metric in assessed_metrics:
#             answer = db.aus_assessed_statements.find_one(
#                 {'metric': metric, 'company': statement['company'], 'year': statement['year']})
#             print(answer)
#             if len(answer['checked_by']) == 0:
#                 verified = False
#                 unverified_answers.append(answer['metric'])
#                 unverified_answers_ids.append(answer['_id'])
#             else:
#                 number_of_verified_answers += 1
#                 verified_answers.append(answer['metric'])
#                 verified_answers_ids.append(answer['_id'])
#         if verified:
#             db.assessed_0523.update_one({'_id': statement['_id']},
#                                         {'$set': {'verified': True,
#                                                   'verified_answers': verified_answers,
#                                                   'unverified_answers': unverified_answers,
#                                                   'verified_answers_ids': verified_answers_ids,
#                                                   'unverified_answers_ids': unverified_answers_ids
#                                                   }})
#         else:
#             db.assessed_0523.update_one({'_id': statement['_id']},
#                                         {'$set': {'number_of_verified_answers': number_of_verified_answers,
#                                                   'verified_answers': verified_answers,
#                                                   'unverified_answers': unverified_answers,
#                                                   'verified_answers_ids': verified_answers_ids,
#                                                   'unverified_answers_ids': unverified_answers_ids
#                                                   }})
#
#
# verify_statements()
#
# assessed_statements = db.assessed_0523.find({'verified': {'$exists': False}})
#
# f = open("D:\\AUS Assessed Stats.csv", "a")
#
# for statement in assessed_statements:
#     f.write("{0};{1};{2};{3};{4};{5};{6}\n".format(
#         statement['company'],
#         statement['year'],
#         statement['number_of_verified_answers'],
#         statement['unverified_answers'],
#         statement['verified_answers'],
#         statement['unverified_answers_ids'],
#         statement['verified_answers_ids']
#     ))
#
# f.close()



class Cursor(object):

    def __init__(self, method, per_page=20, **kwargs):
        self.method = method
        self.kwargs = kwargs
        if per_page > 200:
            self.per_page = 200
        else:
            self.per_page = per_page

        self.offset = 0
        self.limit = per_page
        self.items = None

    def has_next(self) -> bool:
        self.items = self.method(offset=self.offset, limit=self.limit, **self.kwargs)
        return len(self.items) > 0

    def next(self):
        self.offset += self.per_page
        return self.items



import time

import wikirate4py
from pymongo import MongoClient

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("fti_project")

api = wikirate4py.API('rMrVM3sSrtsiX23fKfsM5Qtt')

metrics = db.pc_answers_2023.distinct("metric")

for metric in metrics:
    print(metric)
    fields = metric.split("+")
    answers = api.get_answers(fields[1], fields[0], published="false")
    print(answers)

# for answer in answers:
#     # sources = set(answer['wikirate_sources'])
#     # source = ''
#     # for s in sources:
#     #     source += s + "\n"
#     #
#     # value = ''
#     # if len(answer['value']) > 1:
#     #     for v in answer['value']:
#     #         value += v + "\n"
#     # else:
#     #     value = answer['value'][0]
#
#     metric_name = answer['metric'].split('+')[1]
#     metric_designer = answer['metric'].split('+')[0]
#     # comment = ''
#     # if answer.get('comment') is not None:
#     #     comment = answer['comment']
#
#     for wid in answer['wikirate_id']:
#         start_time = time.time()
#         try:
#             # if 'import_status_' + str(wid) in answer:
#             #     continue
#             print(answer['year'])
#             print(metric_name)
#             print(wid)
#             metric = api.get_answer_by(metric_designer=metric_designer,
#                                        metric_name=metric_name,
#                                        company=wid,
#                                        year=answer['year'])
#             a = api.update_research_metric_answer_by_id(id=metric.id,
#                                                         unpublished="0")
#             end_time = time.time()
#             exec_time = end_time - start_time
#             print("Exec Time: " + exec_time.__str__())
#             print(a)
#             db.answers_2023.update_one({'_id': answer['_id']}, {'$set': {'import_status_' + str(wid): 'imported'}})
#         except Exception as e:
#             print(e.__str__())
#
# client.close()



import re
from setuptools import setup, find_packages

VERSION_FILE = "wikirate4py/__init__.py"
with open(VERSION_FILE) as version_file:
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                      version_file.read(), re.MULTILINE)
with open('README.rst') as readme_file:
    readme = readme_file.read()

if match:
    version = match.group(1)
else:
    raise RuntimeError(f"Unable to find version string in {VERSION_FILE}.")

tests_require = [
    "mock",
    "nose",
    "vcrpy",
    "html2text",
    "python-dotenv"
]
setup(name='wikirate4py',
      version=version,
      description='WikiRate for Python!',
      long_description=readme,
      url='https://github.com/wikirate/wikirate4py',
      author='Vasiliki Gkatziaki',
      author_email='vasso@wikirate.org',
      license='GPL-3.0',
      download_url='https://github.com/wikirate/wikirate4py/archive/refs/tags/v1.2.5.tar.gz',
      packages=find_packages(exclude=["tests", "examples"]),
      install_requires=[
          "requests",
          "html2text",
          "pandas"
      ],
      project_urls={
          "Documentation": "https://wikirate4py.readthedocs.io",
          "Issue Tracker": "https://github.com/wikirate4py/wikirate4py/issues",
          "Source Code": "https://github.com/wikirate4py/wikirate4py",
      },
      extras_require={
          "test": tests_require,
      },
      test_suite="nose.collector",
      keywords="wikirate library",
      python_requires=">=3.6",
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Topic :: Software Development :: Libraries",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3 :: Only",
      ],
      zip_safe=True)



from pymongo import MongoClient

import wikirate4py

client = MongoClient("localhost:27017",
                     username='admin',
                     password='w00d4th3s0u1',
                     authSource='admin')

db = client.get_database("msa_assessed")

api = wikirate4py.API('NzSK5muu3o7oSwDa2PpUVQtt')

statements = db.assessed_0523.find({"verified": True})

for statement in statements:
    print('Statement ID: {0}'.format(statement['_id']))
    unverified_answer_ids: [] = statement['unverified_answers_ids']
    unverified_answers: [] = statement['unverified_answers']
    verified_answer_ids: [] = statement['verified_answers_ids']
    verified_answers: [] = statement['verified_answers']

    for id in unverified_answer_ids.copy():
        print(id)
        answer = api.get_answer(id)
        if len(answer.checked_by) == 0:
            response = api.verify_answer(id)
            print(response)
            checked_by = response.json()
            name = checked_by['name'].split('+')
        else:
            name = answer.metric
            print('already verified: {0}'.format(id))
        unverified_answer_ids.remove(id)
        verified_answer_ids.append(id)
        verified_answers.append('{0}+{1}'.format(name[0], name[1]))
        try:
            unverified_answers.remove('{0}+{1}'.format(name[0], name[1]))
        except ValueError as e:
            print(e.__str__())

    db.assessed_0523.update_one({'_id': statement['_id']}, {
        '$set': {'unverified_answers_ids': unverified_answer_ids,
                 'verified_answers_ids': verified_answer_ids,
                 'verified_answers': verified_answers,
                 'unverified_answers': unverified_answers}})

    print(statement['url'])



import functools
import logging
import os
import sys
import re

import requests
from os import environ
from urllib.parse import (
    urlparse, urljoin
)

from wikirate4py.exceptions import IllegalHttpMethod, BadRequestException, UnauthorizedException, \
    ForbiddenException, NotFoundException, TooManyRequestsException, WikiRateServerErrorException, HTTPException, \
    WikiRate4PyException
from wikirate4py.models import (Company, Topic, Metric, ResearchGroup, CompanyGroup, Source, CompanyItem, MetricItem,
                                Answer, ResearchGroupItem, RelationshipAnswer, SourceItem, TopicItem, AnswerItem,
                                CompanyGroupItem, RelationshipAnswerItem, Region, Project, ProjectItem, RegionItem,
                                Dataset, DatasetItem)

log = logging.getLogger(__name__)

WIKIRATE_API_URL = environ.get(
    'WIKIRATE_API_URL', 'https://wikirate.org/')
WIKIRATE_API_PATH = urlparse(WIKIRATE_API_URL).path


def generate_url_key(input_string):
    # Replace spaces, commas, single quotes, dots, and special characters with underscores, and convert to lowercase
    url_key = re.sub(r'[^a-zA-Z0-9_+]+', lambda x: ' ' if x.group(0) != '+' else '+', input_string)
    # Check and remove double underscores
    url_key = re.sub(r"\s+", "_", url_key)
    return url_key


def objectify(wikirate_obj, list=False):
    def decorator(method):
        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            payload = method(*args, **kwargs).json()
            if not list:
                return wikirate_obj(payload)
            else:
                return [wikirate_obj(item) for item in payload.get("items")]

        return wrapper

    return decorator


class API(object):
    allowed_methods = ['post', 'get', 'delete']
    content_type_specified = True

    def __init__(self, oauth_token, wikirate_api_url=WIKIRATE_API_URL, auth=()):
        self.oauth_token = oauth_token
        self.wikirate_api_url = wikirate_api_url
        self.auth = auth
        self.session = requests.Session()

    @property
    def headers(self):
        headers = {
            'X-API-Key': self.oauth_token
        }
        return headers

    def request(self, method, path, headers, params, files={}):
        method = method.strip().lower()
        if method not in self.allowed_methods:
            msg = "The '{0}' method is not accepted by the WikiRate " \
                  "client.".format(method)
            raise IllegalHttpMethod(msg)

        # if an error was returned throw an exception
        try:
            response = self.session.request(method, path, auth=self.auth, data=params, headers=headers, timeout=480,
                                            files=files)

            if files.get("card[subcards][+file][file]") is not None:
                files.get("card[subcards][+file][file]").close()
        except Exception as e:
            raise WikiRate4PyException(f'Failed to send request: {e}').with_traceback(sys.exc_info()[2])
        finally:
            self.session.close()

        if response.status_code == 400:
            raise BadRequestException(response)
        if response.status_code == 401:
            raise UnauthorizedException(response)
        if response.status_code == 403:
            raise ForbiddenException(response)
        if response.status_code == 404:
            raise NotFoundException(response)
        if response.status_code == 429:
            raise TooManyRequestsException(response)
        if response.status_code >= 500:
            raise WikiRateServerErrorException(response)
        if response.status_code and not 200 <= response.status_code < 300:
            raise HTTPException(response)

        # else return the response
        return response

    def get(self, path, endpoint_params=(), filters=(), **kwargs):
        headers = self.headers
        if 'content-type' in headers:
            headers.pop('content-type')

        params = {}
        for k, arg in kwargs.items():
            if arg is None:
                continue
            if k not in endpoint_params and k not in filters:
                log.warning(f'Unexpected parameter: {k}')
            if k in filters:
                if k == 'value_from' or k == 'value_to':
                    params['filter[value]' + '[' + re.sub(r'.*_', '', k) + ']'] = str(arg)
                elif k in ['subject_company_name', 'object_company_name', 'object_company_id', 'subject_company_id']:
                    params['filter[' + k + '][]'] = arg if isinstance(arg, str) else arg
                else:
                    params['filter[' + k + ']'] = str(arg)
            else:
                params[k] = str(arg)
        log.debug("PARAMS: %r", params)

        # Get the function path
        path = self.format_path(path, self.wikirate_api_url)
        return self.request('get', path, headers=headers, params=params or {})

    def post(self, path, params={}, files={}):
        path = self.format_path(path, self.wikirate_api_url)
        return self.request('post', path, headers=self.headers, params=params or {}, files=files)

    def delete(self, path, params={}):
        path = self.format_path(path, self.wikirate_api_url)
        return self.request('delete', path, headers=self.headers, params=params or {})

    def format_path(self, path, wikirate_api_url=WIKIRATE_API_URL):
        # Probably a webhook path
        if path.startswith(wikirate_api_url):
            return path

        # Using the HTTP shortcut
        if path.startswith("/"):
            return urljoin(wikirate_api_url, path.lstrip('/'))

    def list_to_str(self, list):
        value_str = ''
        for t in list:
            value_str += t + '\n'
        return value_str

    @objectify(Company)
    def get_company(self, identifier) -> Company:
        """get_company(identifier)

        Returns a company based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.Company`
        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(
                identifier.replace(',', ' ').replace('.', ' ').replace('/', ' ').replace('-', ' ').strip().replace(" ",
                                                                                                                   "_")))

    @objectify(CompanyItem, list=True)
    def get_companies(self, **kwargs):
        """get_companies(*, offset, limit)

        Returns a list of WikiRate Companies

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.CompanyItem`]

        """
        return self.get("/Company.json", endpoint_params=('limit', 'offset'),
                        filters=('name', 'company_category', 'company_group', 'country'),
                        **kwargs)

    @objectify(CompanyItem, list=True)
    def get_companies_of(self, identifier, **kwargs):
        """get_companies(*, offset, limit)

        Returns a list of WikiRate Companies

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.CompanyItem`]

        """
        return self.get("/~{0}+companies.json".format(identifier), endpoint_params=('limit', 'offset'),
                        filters=('name', 'company_category', 'company_group', 'country'),
                        **kwargs)

    @objectify(Topic)
    def get_topic(self, identifier):
        """get_topic(identifier)

        Returns a WikiRate Topic based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.Topic`
        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(
                identifier.replace(',', ' ').replace('.', ' ').replace('/', ' ').replace('-', ' ').strip().replace(" ",
                                                                                                                   "_")))

    @objectify(TopicItem, True)
    def get_topics(self, **kwargs):
        """get_topics(*, offset, limit)

        Returns a list of WikiRate Topics

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
        :py:class:`List`\[:class:`~wikirate4py.models.TopicItem`]

        """
        return self.get("/Topics.json", endpoint_params=('limit', 'offset'), filters=('name', 'bookmark'), **kwargs)

    @objectify(Metric)
    def get_metric(self, identifier=None, metric_name=None, metric_designer=None):
        """get_metric(metric_name, metric_designer)

        Returns a WikiRate Metric based on the given metric name and metric designer.

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        metric_name
            name of metric

        metric_designer
            name of metric designer


        Returns
        -------
            :py:class:`~wikirate4py.models.Metric`
        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        elif isinstance(identifier, str):
            return self.get("/{0}.json".format(identifier.replace(" ", "_")))
        else:
            return self.get("/{0}+{1}.json".format(metric_designer.replace(" ", "_"), metric_name.replace(" ", "_")))

    @objectify(MetricItem, list=True)
    def get_metrics(self, **kwargs):
        """get_metrics(*, offset, limit)

        Returns a list of WikiRate Metrics

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.MetricItem`]

        """
        return self.get("/Metrics.json", endpoint_params=('limit', 'offset'), filters=(
            'name', 'bookmark', 'wikirate_topic', 'designer', 'published', 'metric_type', 'value_type',
            'research_policy',
            'dataset'), **kwargs)

    @objectify(MetricItem, list=True)
    def get_metrics_of(self, identifier, **kwargs):
        """get_metrics(*, offset, limit)

        Returns a list of WikiRate Metrics

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.MetricItem`]

        """

        return self.get("/~{0}+Metrics.json".format(identifier), endpoint_params=('limit', 'offset'), filters=(
            'name', 'bookmark', 'wikirate_topic', 'designer', 'published', 'metric_type', 'value_type',
            'research_policy',
            'dataset'), **kwargs)

    @objectify(ResearchGroup)
    def get_research_group(self, identifier):
        """get_research_group(identifier)

        Returns a WikiRate Research Group based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.ResearchGroup`
        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(
                identifier.replace(',', ' ').replace('.', ' ').replace('/', ' ').replace('-', ' ').strip().replace(" ",
                                                                                                                   "_")))

    @objectify(ResearchGroupItem, list=True)
    def get_research_groups(self, **kwargs):
        """get_research_groups(*, offset, limit)

        Returns a list of WikiRate Research Groups

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.ResearchGroupItem`]
        """
        return self.get("/Research_Groups.json", endpoint_params=('limit', 'offset'), filters=('name',), **kwargs)

    @objectify(CompanyGroup)
    def get_company_group(self, identifier):
        """get_company_group(identifier)

        Returns a WikiRate Company Group based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.CompanyGroup`
        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(
                identifier.replace(',', ' ').replace('.', ' ').replace('/', ' ').replace('-', ' ').strip().replace(" ",
                                                                                                                   "_")))

    @objectify(CompanyGroupItem, True)
    def get_company_groups(self, **kwargs):
        """get_company_groups(*, offset, limit)

        Returns a list of WikiRate Company Groups

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.CompanyGroupItem`]

        """
        return self.get("/Company_Groups.json", endpoint_params=('limit', 'offset'), filters=('name',), **kwargs)

    @objectify(Source)
    def get_source(self, identifier):
        """get_source(identifier)

        Returns a WikiRate Source based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.Source`
        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(
                identifier.replace(',', ' ').replace('.', ' ').replace('/', ' ').replace('-', ' ').strip().replace(" ",
                                                                                                                   "_")))

    @objectify(SourceItem, True)
    def get_sources(self, **kwargs):
        """get_sources(*, offset, limit)

        Returns a list of WikiRate Sources

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.
        company_name
            filter sources where the company name matches fully or partially the given string
        year
            filter sources based on given year
        wikirate_title
            filter sources where their title match fully or partially the given string
        report_type
            filter sources based on the report type
        wikirate_topic
            filter sources based on given topic
        wikirate_link
            filter sources where their url matches fully or partially the given string

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.Source`]
        """
        return self.get("/Sources.json", endpoint_params=('limit', 'offset'), filters=(
            'name', 'wikirate_title', 'wikirate_topic', 'report_type', 'year', 'wikirate_link', 'company_name'),
                        **kwargs)

    @objectify(Answer)
    def get_answer(self, identifier):
        """get_answer(identifier)

        Returns a relationship metric answer given its identifier the identifier can be numeric or alphanumeric.

        Parameters
        ----------
        identifier
            numeric or alphanumeric identifier of the relationship metric answer, example of alphanumeric identifier:
            Core+Country+Adidas AG+2019

        Returns
        -------
            :py:class:`~wikirate4py.models.Answer`
        """

        url_key = generate_url_key(identifier) if isinstance(identifier, str) else f"~{identifier}"

        return self.get(f"/{url_key}.json")

    @objectify(AnswerItem, True)
    def get_answers(self, entity, **kwargs):
        """get_answers(identifier, *, offset, limit, year, status, company_group, country, company_id, value, value_from, value_to, \
                       updated, updater, designer, outliers, source, verification, project, bookmark, view)

        Returns a list of WikiRate Answers by entity (it can be metric name/id, dataset name/id, company name/id or source name/id)

        Parameters
        ----------
        entity
            numeric wikirate identifier or alphanumeric wikirate entity name for instance: Adidas_AG,

        offset
            default value 0, the (zero-based) offset of the first item in the collection to return

        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        year
            answer year

        status
            `all`, `exists` (researched), `known`, `unknown`, or `none` (not researched)

        company_group
            company group name, restricts to answers with companies belonging in the specified company group

        country
            country name, restricts to answers with companies located in the specified country

        company_id
            company identifier, restricts to answers of the defined company

        company_name
            restricts to answers of the defined company name

        value
            answer value to match

        value_from
            restricts to answers with value greater than equal the specified value

        value_to
            restricts to answers with value less than equal the specified value

        updated
            `today`, `week` (this week), `month` (this month)

        updater
            - `wikirate_team`, restricts to answers updated by the WikiRate team
            - `current_user`, restricts to answers updated by you

        outliers
            get either `only` answers considered as outliers or get answers after `exclude` the outliers

        source
            source name, restricts to answers citing the specified source

        verification
            restricts to answers mapped to the defined verification level:
                - `steward_added`: answers added by account with "steward" status
                - `flagged`: answers which have been flagged by the Researcher adding the answer to request verification
                - `community_added`: answers added by community members (e.g. students / volunteers)
                - `community_verified`: answers verified by community members (e.g. students / volunteers)
                - `steward_verified`: answers verified by account with "steward" status
                - `current_user`: answers verified by you
                - `wikirate_team`: answers verified by WikiRate team

        project
            project name, restrict to answers connected to the specified WikiRate project

        bookmark
            - `bookmark`, restrict to answers you have bookmarked
            - `nobookmark`, restrict to answers you have not bookmarked

        Returns
        -------
        :py:class:`List`\[:class:`~wikirate4py.models.AnswerItem`]
        """
        url_key = generate_url_key(entity) if isinstance(entity, str) else f"~{entity}"

        return self.get(f"/{url_key}+Answer.json", endpoint_params=('limit', 'offset', 'view'),
                        filters=('year', 'status', 'company_group', 'country', 'value', 'value_from', 'value_to',
                                 'updated', 'company_id', 'company_name', 'dataset', 'updater', 'outliers', 'source',
                                 'verification', 'bookmark', 'published', 'metric_name', 'designer'),
                        **kwargs)

    @objectify(AnswerItem, True)
    def get_metric_answers(self, metric_name, metric_designer, **kwargs):
        """get_metric_answers(metric_name, metric_designer, *, company_name, company_id, offset, limit, year, status,
        company_group, country, value, value_from, value_to, updated, updater, outliers, source, verification, project,
        bookmark)

        Returns a list of WikiRate Answers

        Parameters
        ----------
        metric_name
            name of relationship metric
        metric_designer
            name of relationship metric designer

        company_name
            restrict the answers based on the given company

        company_id
            restrict the answers based on the given company

        offset
            default value 0, the (zero-based) offset of the first item in the collection to return

        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        year
            answer year

        status
            `all`, `exists` (researched), `known`, `unknown`, or `none` (not researched)

        company_group
            company group name, restricts to answers with companies belonging in the specified company group

        country
            country name, restricts to answers with companies located in the specified country

        value
            answer value to match

        value_from
            restricts to answers with value greater than equal the specified value

        value_to
            restricts to answers with value less than equal the specified value

        updated
            `today`, `week` (this week), `month` (this month)

        updater
            - `wikirate_team`, restricts to answers updated by the WikiRate team
            - `current_user`, restricts to answers updated by you

        outliers
            get either `only` answers considered as outliers or get answers after `exclude` the outliers

        source
            source name, restricts to answers citing the specified source

        verification
            restricts to answers mapped to the defined verification level:
                - `steward_added`: answers added by account with "steward" status
                - `flagged`: answers which have been flagged by the Researcher adding the answer to request verification
                - `community_added`: answers added by community members (e.g. students / volunteers)
                - `community_verified`: answers verified by community members (e.g. students / volunteers)
                - `steward_verified`: answers verified by account with "steward" status
                - `current_user`: answers verified by you
                - `wikirate_team`: answers verified by WikiRate team

        project
            project name, restrict to answers connected to the specified WikiRate project

        bookmark
            - `bookmark`, restrict to answers you have bookmarked
            - `nobookmark`, restrict to answers you have not bookmarked
        published
            - `true`, returns only published answers (default mode)
            - `false`, returns only unpublished answers
            - `all`, returns all published and unpublished answers


        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.AnswerItem`]

        """
        return self.get(
            "/{0}+{1}+Answer.json".format(metric_designer.replace(" ", "_"), metric_name.replace(" ", "_")),
            endpoint_params=('limit', 'offset'),
            filters=('year', 'status', 'company_group', 'country', 'value', 'value_from', 'value_to', 'updated',
                     'company_id', 'company_name', 'dataset', 'updater', 'outliers', 'source', 'verification',
                     'bookmark', 'published', 'view'), **kwargs)

    @objectify(RelationshipAnswer)
    def get_relationship_answer(self, identifier):
        """get_relationship_answer(identifier)

        Returns a relationship metric answer given its identifier the identifier can be numeric or alphanumeric.

        Parameters
        ----------
        identifier
            numeric or alphanumeric identifier of the relationship metric answer

        Returns
        -------
            :py:class:`~wikirate4py.models.RelationshipAnswer`
        """

        url_key = generate_url_key(identifier) if isinstance(identifier, str) else f"~{identifier}"

        return self.get(f"/{url_key}.json")

    @objectify(RelationshipAnswerItem, True)
    def get_relationship_answers(self, entity, **kwargs):
        """get_relationship_answers(entity, *, offset, limit, year, status, company_group, country, value, value_from, value_to, \
                       updated, updater, outliers, source, verification, project, bookmark)

        Returns a list of WikiRate Relationship Answers

        Parameters
        ----------
        entity
            numeric or alphanumeric identifier of and entity that can has relationships eg. relationship metrics,
            companies, datasets etc.

        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        year
            answer year

        status
            `all`, `exists` (researched), `known`, `unknown`, or `none` (not researched)

        company_group
            company group name, restricts to relationship answers with subject companies belonging in the specified company group

        country
            country name, restricts to relationship answers with subject companies located in the specified country

        value
            answer value to match

        value_from
            restricts to relationship answers with value greater than equal the specified value

        value_to
            restricts to relationship answers with value less than equal the specified value

        updated
            `today`, `week` (this week), `month` (this month)

        updater
            - `wikirate_team`, restricts to relationship answers updated by the WikiRate team
            - `current_user`, restricts to relationship answers updated by you

        outliers
            get either `only` relationship answers considered as outliers or get answers after `exclude` the outliers

        source
            source name, restricts to answers citing the specified source

        verification
            restricts to relationship answers mapped to the defined verification level:

                - `steward_added`: relationship answers added by account with "steward" status
                - `flagged`: relationship answers which have been flagged by the Researcher adding the answer to request verification
                - `community_added`: relationship answers added by community members (e.g. students / volunteers)
                - `community_verified`: relationship answers verified by community members (e.g. students / volunteers)
                - `steward_verified`: relationship answers verified by account with "steward" status
                - `current_user`: relationship answers verified by you
                - `wikirate_team`: relationship answers verified by WikiRate team

            project
                project name, restrict to relationship answers connected to the specified WikiRate project

            bookmark
                - `bookmark`, restrict to relationship answers you have bookmarked
                - `nobookmark`, restrict to relationship answers you have not bookmarked

            published
            - `true`, returns only published answers (default mode)
            - `false`, returns only unpublished answers
            - `all`, returns all published and unpublished answers

            Returns
            -------
                :py:class:`List`\[:class:`~wikirate4py.models.RelationshipAnswerItem`]

            """

        url_key = generate_url_key(entity) if isinstance(entity, str) else f"~{entity}"

        return self.get(f"/{url_key}+Relationship_Answer.json",
                        endpoint_params=('limit', 'offset'), filters=(
                'year', 'status', 'company_group', 'country', 'value', 'value_from', 'value_to', 'updated',
                'updater', 'outliers', 'source', 'verification', 'project', 'bookmark', 'published',
                'object_company_name', 'subject_company_name', 'object_company_id', 'subject_company_id'), **kwargs)

    @objectify(RelationshipAnswerItem, True)
    def get_relationship_metric_answers(self, metric_name, metric_designer, **kwargs):
        """get_relationship_metric_answers(metric_name, metric_designer, *, offset, limit, year, status, company_group, country, value, value_from, value_to, \
                       updated, updater, outliers, source, verification, project, bookmark)
        Returns a list of WikiRate Relationship Answers

        Parameters
        ----------
        metric_name
            name of relationship metric
        metric_designer
            name of relationship metric designer

        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.
        year
            answer year

        status
            `all`, `exists` (researched), `known`, `unknown`, or `none` (not researched)

        company_group
            company group name, restricts to relationship answers with subject companies belonging in the specified company group

        country
            country name, restricts to relationship answers with subject companies located in the specified country

        value
            answer value to match

        value_from
            restricts to relationship answers with value greater than equal the specified value

        value_to
            restricts to relationship answers with value less than equal the specified value

        updated
            `today`, `week` (this week), `month` (this month)

        updater
            - `wikirate_team`, restricts to relationship answers updated by the WikiRate team
            - `current_user`, restricts to relationship answers updated by you

        outliers
            get either `only` relationship answers considered as outliers or get answers after `exclude` the outliers

        source
            source name, restricts to answers citing the specified source

        verification
            restricts to relationship answers mapped to the defined verification level:
                - `steward_added`: relationship answers added by account with "steward" status
                - `flagged`: relationship answers which have been flagged by the Researcher adding the answer to request verification
                - `community_added`: relationship answers added by community members (e.g. students / volunteers)
                - `community_verified`: relationship answers verified by community members (e.g. students / volunteers)
                - `steward_verified`: relationship answers verified by account with "steward" status
                - `current_user`: relationship answers verified by you
                - `wikirate_team`: relationship answers verified by WikiRate team

            project
                project name, restrict to relationship answers connected to the specified WikiRate project

            bookmark
                - `bookmark`, restrict to relationship answers you have bookmarked
                - `nobookmark`, restrict to relationship answers you have not bookmarked

            published
            - `true`, returns only published answers (default mode)
            - `false`, returns only unpublished answers
            - `all`, returns all published and unpublished answers

            Returns
            -------
                :py:class:`List`\[:class:`~wikirate4py.models.RelationshipAnswerItem`]

            """

        return self.get("/~{0}+{1}+Relationship_Answer.json".format(metric_designer.replace(" ", "_"),
                                                                    metric_name.replace(" ", "_")),
                        endpoint_params=('limit', 'offset', 'view'), filters=(
                'year', 'status', 'company_group', 'country', 'value', 'value_from', 'value_to', 'updated',
                'updater', 'outliers', 'source', 'verification', 'project', 'bookmark', 'published',
                'object_company_name', 'subject_company_name', 'object_company_id', 'subject_company_id'), **kwargs)

    @objectify(Project)
    def get_project(self, identifier):
        """get_project(identifier)
        Returns a WikiRate Project based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.Project`

        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(
                identifier.replace(',', ' ').replace('.', ' ').replace('/', ' ').replace('-', ' ').strip().replace(" ",
                                                                                                                   "_")))

    @objectify(ProjectItem, True)
    def get_projects(self, **kwargs):
        """get_projects(*, offset, limit)

        Returns a list of WikiRate Projects

        Parameters
        -------------------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.ProjectItem`]

        """
        return self.get("/Projects.json", endpoint_params=('limit', 'offset'), filters=('name', 'wikirate_status'),
                        **kwargs)

    @objectify(Dataset)
    def get_dataset(self, identifier):
        """get_dataset(identifier)
        Returns a WikiRate Dataset based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.Dataset`

        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(
                identifier.replace(',', ' ').replace('.', ' ').replace('/', ' ').replace('-', ' ').strip().replace(" ",
                                                                                                                   "_")))

    @objectify(DatasetItem, True)
    def get_datasets(self, **kwargs):
        """get_datasets(*, offset, limit)

        Returns a list of WikiRate Datasets

        Parameters
        -------------------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.DatasetItem`]

        """
        return self.get("/Data_sets.json", endpoint_params=('limit', 'offset'), filters=('name', 'wikirate_topic'),
                        **kwargs)

    @objectify(RegionItem, True)
    def get_regions(self, **kwargs):
        """get_regions(*, offset, limit)

        Returns the list of all geographic regions we use in WikiRate platform

        Parameters
        ----------
        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.Region`]

        """
        return self.get("/Region.json", endpoint_params=('limit', 'offset'), **kwargs)

    @objectify(Region)
    def get_region(self, identifier):
        """get_project(identifier)
        Returns a WikiRate Region based on the given identifier (name or number)

        Parameters
        ----------
        identifier
            two different identifiers are allowed for WikiRate entities, numerical identifiers or name identifiers

        Returns
        -------
            :py:class:`~wikirate4py.models.Project`

        """
        if isinstance(identifier, int):
            return self.get("/~{0}.json".format(identifier))
        else:
            return self.get("/{0}.json".format(identifier.replace(" ", "_")))

    def search_by_name(self, entity, name, **kwargs):
        """search_by_name(entity, name, *, offset, limit)

        Searches for a Company or Metric or Topic or Company Group or Research Group or Project by a given name.
        If offset and limit are not defined it returns the first 20 search results.

        Parameters
        -------------------
        entity
            allowed entities:
                - :py:class:`~wikirate4py.models.Company`,
                - :py:class:`~wikirate4py.models.Metric`,
                - :py:class:`~wikirate4py.models.Topic`,
                - :py:class:`~wikirate4py.models.CompanyGroup`,
                - :py:class:`~wikirate4py.models.ResearchGroup`,
                - :py:class:`~wikirate4py.models.Project`
        name
            search term

        offset
            default value 0, the (zero-based) offset of the first item in the collection to return

        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------

            - A :py:class:`List` of :py:class:`~wikirate4py.models.CompanyItem`, or
            - A :py:class:`List` of :py:class:`~wikirate4py.models.MetricItem`, or
            - A :py:class:`List` of :py:class:`~wikirate4py.models.TopicItem`, or
            - A :py:class:`List` of :py:class:`~wikirate4py.models.CompanyGroupItem`, or
            - A :py:class:`List` of :py:class:`~wikirate4py.models.ResearchGroupItem`, or
            - A :py:class:`List` of :py:class:`~wikirate4py.models.ProjectItem`

        """
        if entity is Company:
            return self.get_companies(name=name, **kwargs)
        elif entity is Metric:
            return self.get_metrics(name=name, **kwargs)
        elif entity is Topic:
            return self.get_topics(name=name, **kwargs)
        elif entity is CompanyGroup:
            return self.get_company_groups(name=name, **kwargs)
        elif entity is ResearchGroup:
            return self.get_research_groups(name=name, **kwargs)
        elif entity is Project:
            return self.get_projects(name=name, **kwargs)
        else:
            raise WikiRate4PyException(f"Type of parameter 'entity' ({type(entity)}) is not allowed")

    @objectify(SourceItem, True)
    def search_source_by_url(self, url, **kwargs):
        """search_source_by_url(url, *, offset, limit)

        Searches for a WikiRare Source based on a given url.
        If offset and limit are not defined it returns the first 20 search results.

        Parameters
        -------------------
        url
            as a search term

        offset
            default value 0, the (zero-based) offset of the first item in the collection to return
        limit
            default value 20, the maximum number of entries to return. If the value exceeds the maximum, then the maximum value will be used.

        Returns
        -------
            :py:class:`List`\[:class:`~wikirate4py.models.Source`]

                """
        kwargs = {
            "query[url]": url
        }
        return self.get("/Source_by_url.json",
                        endpoint_params=('query[url]', 'limit', 'offset'),
                        filters=(),
                        **kwargs)

    @objectify(Company)
    def add_company(self, name, headquarters, **kwargs):
        """add_company(name, headquarters, *, os_id, open_corporates)

        Creates and Returns a company given the company name and headquarters

        Parameters
        ----------
        name
            company name
        headquarters
            name of the region the headquarters of the company is located
        wikipedia
            company's wikipedia page url
        os_id
            company's identifier on https://opensupplyhub.org
        open_corporates
            company's identifier on https://opencorporates.com/

        Returns
        -------
            :py:class:`~wikirate4py.models.Company`

        """

        if name is None or headquarters is None:
            raise WikiRate4PyException(
                'A WikiRate company is defined by a name and headquarters, please be sure you '
                'have defined both while trying to create a new company')
        optional_params = ('os_id', 'wikipedia', 'open_corporates', 'website')
        params = {
            "card[type]": "Company",
            "card[name]": name,
            "card[subcards][+headquarters]": headquarters,
            "confirmed": "true",
            "format": "json",
            "success[format]": "json"
        }

        for k, arg in kwargs.items():
            if arg is None:
                continue
            if k not in optional_params:
                log.warning(f'Unexpected parameter: {k}')
            else:
                params['card[subcards][+' + k + ']'] = str(arg)
        log.debug("PARAMS: %r", params)
        if 'open_corporates' not in kwargs:
            params['card[skip]'] = "update_oc_mapping_due_to_headquarters_entry"

        return self.post("/card/create", params)

    @objectify(Company, False)
    def update_company(self, identifier, **kwargs):
        if identifier is None:
            raise WikiRate4PyException(
                'A Wikirate company is defined by identifier. Please when you try to update a Wikirate company, provide a valid company identifier or name. ')
        optional_params = ('headquarters', 'os_id', 'wikipedia', 'open_corporates', 'sec_cik', 'website')
        params = {
            "card[type]": "Company",
            "format": "json",
            "success[format]": "json"
        }

        for k, arg in kwargs.items():
            if arg is None:
                continue
            if k not in optional_params:
                log.warning(f'Unexpected parameter: {k}')
            else:
                params['card[subcards][+' + k + ']'] = str(arg)
        log.debug("PARAMS: %r", params)

        url_key = generate_url_key(identifier) if isinstance(identifier, str) else f"~{identifier}"

        return self.post(f"/update/{url_key}".format(identifier), params)

    @objectify(Answer)
    def add_research_metric_answer(self, **kwargs):
        """add_research_metric_answer(metric_designer, metric_name, company, year, value, source, *, comment)

        Creates and Returns a company given the company name and headquarters

        Parameters
        ----------
        metric_name
            name of metric

        metric_designer
            name of metric designer

        company
            company the answer is assigned to

        year
            reporting year

        value
            value of the answer

        source
            source name

        comment
            comment on the imported metric answer

        Returns
        -------
            :py:class:`~wikirate4py.models.Answer`

        """
        required_params = ('metric_designer', 'metric_name', 'company', 'year', 'value', 'source')
        optional_params = ('comment', 'unpublished')

        for k in required_params:
            if k not in kwargs:
                raise WikiRate4PyException("""Invalid set of params! You need to define all the following params to import 
                    a new research answer: """ + required_params.__str__())

        company_identifier = '~' + str(kwargs['company']) if isinstance(kwargs['company'], int) else kwargs[
            'company'].replace(' ', '_')
        params = {
            "card[type]": "Answer",
            "card[name]": kwargs['metric_designer'] + '+' + kwargs['metric_name'] + '+' + company_identifier + '+' +
                          str(kwargs['year']),
            "card[subcards][+:value]": kwargs['value'],
            "card[subcards][+:source]": kwargs['source'],
            "format": "json",
            "success[format]": "json"
        }
        for k in kwargs.keys():
            if k == 'comment':
                params['card[subcards][+:discussion]'] = str(kwargs[k])
            elif k != required_params and k in optional_params:
                params['card[subcards][+:' + k + ']'] = kwargs[k]
        log.debug("PARAMS: %r", params)

        return self.post("/card/create", params)

    @objectify(Answer)
    def update_research_metric_answer(self, **kwargs):
        """update_research_metric_answer(metric_designer, metric_name, company, year, *, value, source, comment)

        Updates and Returns an existing metric answer

        Parameters
        ----------
        metric_name
            name of metric

        metric_designer
            name of metric designer

        company
            company the answer is assigned to

        year
            reporting year

        value
            new value

        source
            new source name

        comment
            new comment on the metric answer

        Returns
        -------
            :py:class:`~wikirate4py.models.Answer`

        """
        required_params = ('metric_designer', 'metric_name', 'company', 'year')
        optional_params = ('value', 'source', 'comment', 'unpublished')

        for k in required_params:
            if k not in kwargs or kwargs.get(k) is None:
                raise WikiRate4PyException("""Invalid set of params! You need to define all the following params to import 
                        a new research answer: """ + required_params.__str__())

        company_identifier = '~' + str(kwargs['company']) if isinstance(kwargs['company'], int) else kwargs[
            'company'].replace(' ', '_')
        params = {
            "card[type]": "Answer",
            "card[name]": kwargs['metric_designer'] + '+' + kwargs['metric_name'] + '+' + company_identifier + '+' +
                          str(kwargs['year']),
            "format": "json",
            "success[format]": "json"
        }
        for k in kwargs.keys():
            if k == 'comment':
                params['card[subcards][+:discussion]'] = str(kwargs[k])
            elif k not in required_params and k in optional_params:
                params['card[subcards][+:' + k + ']'] = str(kwargs[k])

        log.debug("PARAMS: %r", params)

        return self.post("/card/update", params)

    @objectify(Answer)
    def update_research_metric_answer_by_id(self, **kwargs):
        """update_research_metric_answer(id, *, company, year, value, source, comment)

        Updates and Returns an existing metric answer

        Parameters
        ----------
        id
            answer's id

        metric_name
            name of metric

        metric_designer
            name of metric designer

        company
            company the answer is assigned to

        year
            reporting year

        value
            new value

        source
            new source name

        comment
            new comment on the metric answer

        Returns
        -------
            :py:class:`~wikirate4py.models.Answer`

        """
        required_param = 'id'
        optional_params = ('company', 'year', 'value', 'source', 'comment', 'unpublished')

        if required_param not in kwargs:
            raise WikiRate4PyException(
                """Invalid set of params! You need to define all the following params to update the research answer: """ + required_param.__str__())

        if kwargs.get('company') is not None:
            company_identifier = '~' + str(kwargs['company']) if isinstance(kwargs['company'], int) else kwargs[
                'company'].replace(' ', '_')
        params = {
            "card[type]": "Answer",
            "card[name]": '~' + kwargs['id'].__str__(),
            "format": "json",
            "success[format]": "json"
        }
        for k in kwargs.keys():
            if k == 'comment':
                params['card[subcards][+:discussion]'] = str(kwargs[k])
            elif k != required_param and k in optional_params:
                params['card[subcards][+:' + k + ']'] = str(kwargs[k])

        log.debug("PARAMS: %r", params)

        return self.post("/card/update", params)

    @objectify(RelationshipAnswer)
    def add_relationship_metric_answer(self, **kwargs):
        """add_relationship_metric_answer(metric_designer, metric_name, company, year, value, source, *, comment)

        Adds and Returns a relationship metric answer

        Parameters
        ----------
        metric_name
            name of metric

        metric_designer
            name of metric designer

        subject_company
            subject company name of the updated answer

        object_company
            object company name of the updated answer

        year
            reporting year

        value
            value of the relationship metric answer

        source
            source name

        comment
            comment on the imported metric answer

        Returns
        -------
            :py:class:`~wikirate4py.models.RelationshipAnswer`

        """
        required_params = (
            'metric_designer', 'metric_name', 'subject_company', 'object_company', 'year', 'value', 'source')

        for k in required_params:
            if k not in kwargs:
                raise WikiRate4PyException("""Invalid set of params! You need to define all the following params to import 
                        a new research answer: """ + required_params.__str__())

        subject_company_identifier = '~' + str(kwargs['subject_company']) \
            if isinstance(kwargs['subject_company'], int) else kwargs['subject_company'].replace(' ', '_')
        object_company_identifier = '~' + str(kwargs['object_company']) \
            if isinstance(kwargs['object_company'], int) else kwargs['object_company'].replace(' ', '_')
        params = {
            "card[type]": "Relationship_Answer",
            "card[name]": kwargs['metric_designer'].replace(" ", "_") + '+' + kwargs[
                'metric_name'].replace(" ", "_") + '+' + subject_company_identifier +
                          '+' + str(kwargs['year']) + '+' + object_company_identifier,
            "card[subcards][+:value]": kwargs['value'],
            "card[subcards][+:source]": kwargs['source'],
            "format": "json",
            "success[format]": "json"
        }
        if kwargs.get('comment') is not None:
            params['card[subcards][+:discussion]'] = str(kwargs['comment'])
        log.debug("PARAMS: %r", params)

        return self.post("/card/create", params)

    @objectify(Metric)
    def add_metric(self, **kwargs):
        """add_metric(designer, name, question, about, methodology, topics, value_type, options, research_policy, report_type)

        Creates and Returns a new Metric

        Parameters
        -------------------
        designer
            designer of the metric

        name
            name of the metric

        question
            the question that needs answering

        about
            about the metric (plain text/html can be given as input)

        methodology
            metric's methodology (plain text/html can be given as input)

        topics
            a list of metrics

        value_type
            value type

        unit
            unit can be given (e.g. USD, EURO, Square meters etc)

        value_options
            a list of value options (if the value type is Multi-Category)

        research_policy
            research policy's options: Community Assessed, Designer Assessed

        report_type
            report type

        Returns
        -------
            :py:class:`~wikirate4py.models.Metric`

        """
        required_params = ['designer', 'name', 'metric_type', 'value_type']
        optional_params = (
            'question', 'about', 'methodology', 'unit', 'topics', 'value_options', 'research_policy', 'report_type')

        for k in required_params:
            if k not in kwargs:
                raise WikiRate4PyException("""Invalid set of params! You need to define all the following params to create
                                a new metric in WikiRate platform: """ + required_params.__str__())

        params = {
            "card[type]": "Metric",
            "card[name]": kwargs['designer'] + '+' + kwargs['name'],
            "card[subcards][+value_type]": kwargs['value_type'],
            "card[subcards][+*metric_type]": kwargs['metric_type'],
            "card[skip]": "requirements",
            "format": "json",
            "success[format]": "json"
        }

        for k in optional_params:
            if k in kwargs.keys():
                if k in ['topics', 'value_options']:
                    params['card[subcards][+' + k + ']'] = self.list_to_str(kwargs[k])
                else:
                    params['card[subcards][+' + k + ']'] = str(kwargs[k])

        log.debug("PARAMS: %r", params)
        return self.post("/card/create", params=params)

    @objectify(RelationshipAnswer)
    def update_relationship_metric_answer(self, **kwargs):
        """update_relationship_metric_answer(metric_designer, metric_name, company, year, value, source)

        Updates and Returns an existing relationship metric answer

        Parameters
        ----------

        metric_name
            name of metric

        metric_designer
            name of metric designer

        subject_company
            subject company name of the updated answer

        object_company
            object company name of the updated answer

        year
            reporting year

        value
            new value

        source
            new source name

        comment
            new comment on the metric answer

        Returns
        -------
            :py:class:`~wikirate4py.models.RelationshipAnswer`

        """
        required_params = ('metric_designer', 'metric_name', 'subject_company', 'year', 'object_company')
        optional_params = ('value', 'source', 'comment')

        for k in required_params:
            if k not in kwargs or kwargs.get(k) is None:
                raise WikiRate4PyException("""Invalid set of params! You need to define all the following params to import 
                            a new research answer: """ + required_params.__str__())

        subject_company_identifier = '~' + str(kwargs['subject_company']) \
            if isinstance(kwargs['subject_company'], int) else kwargs['subject_company'].replace(' ', '_')
        object_company_identifier = '~' + str(kwargs['object_company']) \
            if isinstance(kwargs['object_company'], int) else kwargs['object_company'].replace(' ', '_')
        params = {
            "card[type]": "Relationship_Answer",
            "card[name]": kwargs['metric_designer'].replace(" ", "_") + '+' + kwargs[
                'metric_name'].replace(" ", "_") + '+' + subject_company_identifier +
                          '+' + str(kwargs['year']) + '+' + object_company_identifier,
            "format": "json",
            "success[format]": "json"
        }

        for k in kwargs.keys():
            if k == 'comment':
                params['card[subcards][+:discussion]'] = str(kwargs[k])
            elif k not in required_params and k in optional_params:
                params['card[subcards][+:' + k + ']'] = str(kwargs[k])

        log.debug("PARAMS: %r", params)

        return self.post("/card/update", params)

    @objectify(Source)
    def add_source(self, **kwargs):
        """add_source(link, file, title, company, report_type, year)

        Updates and Returns an existing relationship metric answer

        Parameters
        -------------------
        link
            url of the original source

        file
           file path to the file you want to upload as a source

        title
            give a title on the source

        company
            comment on the imported metric answer

        report_type
            source name

        year
            reporting year
        file
            filepath on the file you want to upload

        Returns
        -------
            :py:class:`~wikirate4py.models.Source`

        """
        required_params = ['title']
        optional_params = ('link', 'company', 'report_type', 'year', 'file')

        for k in required_params:
            if k not in kwargs:
                raise WikiRate4PyException("""Invalid set of params! You need to define all the following params to import 
                            a new source in WikiRate platform: """ + required_params.__str__())
        if 'link' not in kwargs and 'file' not in kwargs:
            raise WikiRate4PyException("""Invalid set of params! You need to define either a link or give a file path to 
                                            upload a file while creating a new source: """ + required_params.__str__())

        params = {
            "card[type]": "Source",
            "card[subcards][+title]": kwargs['title'],
            "card[skip]": "requirements",
            "format": "json",
            "success[format]": "json"
        }
        files = {}
        for k in kwargs.keys():
            if k in optional_params and k == "file":
                data_file = open(os.path.realpath(kwargs[k]), 'rb')
                files["card[subcards][+file][file]"] = data_file
            else:
                if k == 'company' and isinstance(kwargs[k], int):
                    params['card[subcards][+' + k + ']'] = f"~{kwargs[k]}"
                else:
                    params['card[subcards][+' + k + ']'] = str(kwargs[k])
        log.debug("PARAMS: %r", params)
        return self.post("/card/create", params=params, files=files)

    @objectify(Source)
    def upload_file(self, source, file):
        data_file = open(os.path.realpath(file), 'rb')
        params = {
            "format": "json",
            "success[format]": "json"
        }
        files = {"card[subcards][+file][file]": data_file}

        return self.post("/update/{0}".format(source), params=params, files=files)

    @objectify(Source)
    def update_source(self, **kwargs):
        """update_source(name, title, company, report_type, year)

        Updates and Returns an existing source

        Parameters
        ----------
        name
            source name

        title
            updated source title

        company
            updated reporting company

        report_type
            updated report type

        year
            updated reporting year

        Returns
        -------
            :py:class:`~wikirate4py.models.Source`

              """
        required_params = ('name',)
        optional_params = ('company', 'report_type', 'year', 'title')

        for k in required_params:
            if k not in kwargs:
                raise WikiRate4PyException("""Invalid set of params! You need to define all the following params to import 
                            a new source in WikiRate platform: """ + required_params.__str__())

        params = {
            "card[type]": "Source",
            "card[name]": kwargs['name'],
            "card[skip]": "requirements",
            "format": "json",
            "success[format]": "json"
        }
        for k in kwargs.keys():
            if k in optional_params and k != 'url':
                params['card[subcards][+' + k + ']'] = str(kwargs[k])
        log.debug("PARAMS: %r", params)

        return self.post("/card/update", params)

    def delete_wikirate_entity(self, id):
        """delete_wikirate_entity(id)

        Deletes a WikiRate entity based on the given numeric identifier
        """
        return self.delete("/~{0}".format(id))

    def add_companies_to_group(self, group_id, list=[]):
        ids = ""
        for item in list:
            ids += '~[[' + item + ']]\n'
        params = {
            "card[type]": "List",
            "card[name]": '~' + group_id + '+' + 'Company',
            "card[content]": ids,
            "format": "json",
            "success[format]": "json"
        }

        return self.post("/card/update", params)

    def verify_answer(self, identifier):
        params = {
            "card[type]": "List",
            "card[name]": '~{0}+checked_by'.format(identifier),
            "card[trigger]": 'add_check',
            "format": "json",
            "success[format]": "json"
        }

        return self.post("/card/update", params)

    def get_comments(self, identifier):
        return self.get("/~{0}+discussion.json".format(identifier)).json().get('content', '')

    def get_content(self, identifier):
        return self.get("/{0}.json".format(identifier)).json().get('content', '')



from pandas import DataFrame

from wikirate4py.mixins import WikiRateEntity


def to_dataframe(data):
    if not isinstance(data, list):
        if isinstance(data, WikiRateEntity):
            array = [data.json()]
            return DataFrame.from_dict(array)
        else:
            raise BaseException("""Invalid Input! Provide as input a WikiRateEntity or a list of 
                                            WikiRateEntity objects!""")
    else:
        array = []
        for snippet in data:
            if isinstance(snippet, WikiRateEntity):
                array.append(snippet.json())
            else:
                raise BaseException("""Invalid Input! Provide as input a WikiRateEntity or a list of 
                                                WikiRateEntity objects!""")
        return DataFrame.from_dict(array)